from pathlib import Path

import pandas
import xarray

from cmip_ref_core.constraints import (
    AddSupplementaryDataset,
    RequireContiguousTimerange,
    RequireFacets,
    RequireOverlappingTimerange,
)
from cmip_ref_core.datasets import FacetFilter, SourceDatasetType
from cmip_ref_core.metrics import DataRequirement
from cmip_ref_metrics_esmvaltool.metrics.base import ESMValToolMetric
from cmip_ref_metrics_esmvaltool.recipe import dataframe_to_recipe
from cmip_ref_metrics_esmvaltool.types import OutputBundle, Recipe


class TransientClimateResponse(ESMValToolMetric):
    """
    Calculate the global mean transient climate response for a dataset.
    """

    name = "Transient Climate Response"
    slug = "esmvaltool-transient-climate-response"
    base_recipe = "recipe_tcr.yml"

    experiments = (
        "1pctCO2",
        "piControl",
    )
    data_requirements = (
        DataRequirement(
            source_type=SourceDatasetType.CMIP6,
            filters=(
                FacetFilter(
                    facets={
                        "variable_id": ("tas",),
                        "experiment_id": experiments,
                    },
                ),
            ),
            group_by=("source_id", "member_id", "grid_label"),
            constraints=(
                RequireFacets("experiment_id", experiments),
                RequireContiguousTimerange(group_by=("instance_id",)),
                RequireOverlappingTimerange(group_by=("instance_id",)),
                AddSupplementaryDataset.from_defaults("areacella", SourceDatasetType.CMIP6),
            ),
        ),
    )

    @staticmethod
    def update_recipe(recipe: Recipe, input_files: pandas.DataFrame) -> None:
        """Update the recipe."""
        # Only run the diagnostic that computes TCR for a single model.
        recipe["diagnostics"] = {
            "cmip6": {
                "description": "Calculate TCR.",
                "variables": {
                    "tas": {
                        "preprocessor": "spatial_mean",
                    },
                },
                "scripts": {
                    "tcr": {
                        "script": "climate_metrics/tcr.py",
                        "calculate_mmm": False,
                    },
                },
            },
        }

        # Prepare updated datasets section in recipe. It contains two
        # datasets, one for the "1pctCO2" and one for the "piControl"
        # experiment.
        recipe_variables = dataframe_to_recipe(input_files)
        recipe_variables = {k: v for k, v in recipe_variables.items() if k != "areacella"}

        # Select a timerange covered by all datasets.
        start_times, end_times = [], []
        for variable in recipe_variables.values():
            for dataset in variable["additional_datasets"]:
                start, end = dataset["timerange"].split("/")
                start_times.append(start)
                end_times.append(end)
        timerange = f"{max(start_times)}/{min(end_times)}"

        datasets = recipe_variables["tas"]["additional_datasets"]
        for dataset in datasets:
            dataset["timerange"] = timerange

        recipe["datasets"] = datasets

    @staticmethod
    def format_result(result_dir: Path) -> OutputBundle:
        """Format the result."""
        tcr_file = result_dir / "work/cmip6/tcr/tcr.nc"
        tcr = xarray.open_dataset(tcr_file)

        source_id = tcr.dataset.values[0].decode("utf-8")
        cmec_output = {
            "DIMENSIONS": {
                "model": {source_id: {}},
                "region": {"global": {}},
                "metric": {"tcr": {}},
                "json_structure": [
                    "model",
                    "region",
                    "metric",
                ],
            },
            "RESULTS": {
                source_id: {"global": {"tcr": float(tcr.tcr.values[0])}},
            },
        }

        return cmec_output



import cmip_ref_metrics_esmvaltool
import pytest

import cmip_ref_core.providers
from cmip_ref_core.datasets import DatasetCollection, MetricDataset, SourceDatasetType


@pytest.fixture
def metric_dataset(cmip6_data_catalog) -> MetricDataset:
    selected_dataset = cmip6_data_catalog[
        cmip6_data_catalog["instance_id"] == cmip6_data_catalog.instance_id.iloc[0]
    ]
    return MetricDataset(
        {
            SourceDatasetType.CMIP6: DatasetCollection(
                selected_dataset,
                "instance_id",
            )
        }
    )


def test_example_metric(mocker, tmp_path, metric_dataset, cmip6_data_catalog, definition_factory):
    provider = cmip_ref_metrics_esmvaltool.provider
    provider.prefix = tmp_path / "conda"
    provider.prefix.mkdir()
    provider._conda_exe = provider.prefix / "mock_micromamba"
    provider._conda_exe.touch()
    provider.env_path.mkdir()

    metric = next(
        metric for metric in provider.metrics() if metric.slug == "esmvaltool-global-mean-timeseries"
    )
    ds = cmip6_data_catalog.groupby("instance_id", as_index=False).first()

    definition = definition_factory(cmip6=DatasetCollection(ds, "instance_id"))
    definition.output_directory.mkdir(parents=True)

    result_dir = definition.output_directory / "results" / "recipe_test_a"
    result = result_dir / "work" / "timeseries" / "script1" / "result.nc"

    def mock_run_fn(cmd, *args, **kwargs):
        result.parent.mkdir(parents=True)
        result.touch()

    mock_run = mocker.patch.object(
        cmip_ref_core.providers.subprocess,
        "run",
        autospec=True,
        spec_set=True,
        side_effect=mock_run_fn,
    )
    open_dataset = mocker.patch.object(
        cmip_ref_metrics_esmvaltool.metrics.example.xarray,
        "open_dataset",
        autospec=True,
        spec_set=True,
    )
    open_dataset.return_value.attrs.__getitem__.return_value = "ABC"

    result = metric.run(definition)

    mock_run.assert_called_with(
        [
            f"{provider._conda_exe}",
            "run",
            "--prefix",
            f"{provider.env_path}",
            "esmvaltool",
            "run",
            f"--config-dir={definition.to_output_path('config')}",
            f"{definition.to_output_path('recipe.yml')}",
        ],
        check=True,
    )

    output_bundle_path = definition.output_directory / result.output_bundle_filename
    metric_bundle_path = definition.output_directory / result.metric_bundle_filename

    assert result.successful
    assert output_bundle_path.exists()
    assert output_bundle_path.is_file()
    assert result.output_bundle_filename.name == "output.json"

    assert metric_bundle_path.exists()
    assert metric_bundle_path.is_file()
    assert result.metric_bundle_filename.name == "metric.json"



from __future__ import annotations

import importlib.resources
from pathlib import Path
from typing import TYPE_CHECKING, Any

import pooch
from ruamel.yaml import YAML

from cmip_ref_metrics_esmvaltool.types import Recipe

if TYPE_CHECKING:
    import pandas as pd

yaml = YAML()

FACETS = {
    "CMIP6": {
        "activity": "activity_id",
        "dataset": "source_id",
        "ensemble": "member_id",
        "institute": "institution_id",
        "exp": "experiment_id",
        "grid": "grid_label",
        "mip": "table_id",
        "short_name": "variable_id",
    },
}


def as_isodate(timestamp: pd.Timestamp) -> str:
    """Format a timestamp as an ISO 8601 datetime.

    For example, '2014-12-16 12:00:00' will be formatted as '20141216T120000'.

    Parameters
    ----------
    timestamp
        The timestamp to format.

    """
    return str(timestamp).replace(" ", "T").replace("-", "").replace(":", "")


def as_timerange(group: pd.DataFrame) -> str | None:
    """Format the timeranges from a dataframe as an ESMValTool timerange.

    Parameters
    ----------
    group
        The dataframe describing a single dataset.

    Returns
    -------
        A timerange.
    """
    start_times = group.start_time.dropna()
    if start_times.empty:
        return None
    end_times = group.end_time.dropna()
    if end_times.empty:
        return None  # pragma: no cover
    return f"{as_isodate(start_times.min())}/{as_isodate(end_times.max())}"


def as_facets(
    group: pd.DataFrame,
) -> dict[str, Any]:
    """Convert a group from the datasets dataframe to ESMValTool facets.

    Parameters
    ----------
    group:
        A group of datasets representing a single instance_id.

    Returns
    -------
        A :obj:`dict` containing facet-value pairs.

    """
    facets = {}
    first_row = group.iloc[0]
    project = first_row.instance_id.split(".", 2)[0]
    facets["project"] = project
    for esmvaltool_name, ref_name in FACETS[project].items():
        facets[esmvaltool_name] = getattr(first_row, ref_name)
    timerange = as_timerange(group)
    if timerange is not None:
        facets["timerange"] = timerange
    return facets


def dataframe_to_recipe(files: pd.DataFrame) -> dict[str, Any]:
    """Convert the datasets dataframe to a recipe "variables" section.

    Parameters
    ----------
    files
        The pandas dataframe describing the input files.

    Returns
    -------
        A "variables" section that can be used in an ESMValTool recipe.
    """
    variables: dict[str, Any] = {}
    # TODO: refine to make it possible to combine historical and scenario runs.
    for _, group in files.groupby("instance_id"):
        facets = as_facets(group)
        short_name = facets.pop("short_name")
        if short_name not in variables:
            variables[short_name] = {"additional_datasets": []}
        variables[short_name]["additional_datasets"].append(facets)
    return variables


_ESMVALTOOL_VERSION = "2.11.0"

_RECIPES = pooch.create(
    path=pooch.os_cache("cmip_ref_metrics_esmvaltool"),
    base_url="https://raw.githubusercontent.com/ESMValGroup/ESMValTool/refs/tags/v{version}/esmvaltool/recipes/",
    version=_ESMVALTOOL_VERSION,
    env="REF_METRICS_ESMVALTOOL_DATA_DIR",
)
_RECIPES.load_registry(importlib.resources.open_binary("cmip_ref_metrics_esmvaltool", "recipes.txt"))


def load_recipe(recipe: str) -> Recipe:
    """Load a recipe.

    Parameters
    ----------
    recipe
        The name of an ESMValTool recipe.

    Returns
    -------
        The loaded recipe.
    """
    filename = _RECIPES.fetch(recipe)
    return yaml.load(Path(filename).read_text(encoding="utf-8"))  # type: ignore[no-any-return]


def prepare_climate_data(datasets: pd.DataFrame, climate_data_dir: Path) -> None:
    """Symlink the input files from the Pandas dataframe into a directory tree.

    This ensures that ESMValTool can find the data and only uses the
    requested data.

    Parameters
    ----------
    datasets
        The pandas dataframe describing the input datasets.
    climate_data_dir
        The directory where ESMValTool should look for input data.
    """
    for row in datasets.itertuples():
        if not isinstance(row.instance_id, str):  # pragma: no branch
            msg = f"Invalid instance_id encountered in {row}"
            raise ValueError(msg)
        if not isinstance(row.path, str):  # pragma: no branch
            msg = f"Invalid path encountered in {row}"
            raise ValueError(msg)
        tgt = climate_data_dir.joinpath(*row.instance_id.split(".")) / Path(row.path).name
        tgt.parent.mkdir(parents=True, exist_ok=True)
        tgt.symlink_to(row.path)



"""ESMValTool metrics."""

from cmip_ref_metrics_esmvaltool.metrics.ecs import EquilibriumClimateSensitivity
from cmip_ref_metrics_esmvaltool.metrics.example import GlobalMeanTimeseries
from cmip_ref_metrics_esmvaltool.metrics.tcr import TransientClimateResponse

__all__ = [
    "EquilibriumClimateSensitivity",
    "GlobalMeanTimeseries",
    "TransientClimateResponse",
]



from pathlib import Path

import pandas
import xarray

from cmip_ref_core.constraints import AddSupplementaryDataset, RequireContiguousTimerange
from cmip_ref_core.datasets import FacetFilter, SourceDatasetType
from cmip_ref_core.metrics import DataRequirement
from cmip_ref_metrics_esmvaltool.metrics.base import ESMValToolMetric
from cmip_ref_metrics_esmvaltool.recipe import dataframe_to_recipe
from cmip_ref_metrics_esmvaltool.types import OutputBundle, Recipe


class GlobalMeanTimeseries(ESMValToolMetric):
    """
    Calculate the annual mean global mean timeseries for a dataset.
    """

    name = "Global Mean Timeseries"
    slug = "esmvaltool-global-mean-timeseries"
    base_recipe = "examples/recipe_python.yml"

    data_requirements = (
        DataRequirement(
            source_type=SourceDatasetType.CMIP6,
            filters=(FacetFilter(facets={"variable_id": ("tas",)}),),
            group_by=("instance_id",),
            constraints=(
                RequireContiguousTimerange(group_by=("instance_id",)),
                AddSupplementaryDataset.from_defaults("areacella", SourceDatasetType.CMIP6),
            ),
        ),
    )

    @staticmethod
    def update_recipe(recipe: Recipe, input_files: pandas.DataFrame) -> None:
        """Update the recipe."""
        # Clear unwanted elements from the recipe.
        recipe["datasets"].clear()
        recipe["diagnostics"].pop("map")
        variables = recipe["diagnostics"]["timeseries"]["variables"]
        variables.clear()

        # Prepare updated variables section in recipe.
        recipe_variables = dataframe_to_recipe(input_files)
        recipe_variables = {k: v for k, v in recipe_variables.items() if k != "areacella"}
        for variable in recipe_variables.values():
            variable["preprocessor"] = "annual_mean_global"
            variable["caption"] = "Annual global mean {long_name} according to {dataset}."

        # Populate recipe with new variables/datasets.
        variables.update(recipe_variables)

    @staticmethod
    def format_result(result_dir: Path) -> OutputBundle:
        """Format the result."""
        result = next(result_dir.glob("work/timeseries/script1/*.nc"))
        dataset = xarray.open_dataset(result)

        # TODO: Check how timeseries data are generally serialised
        cmec_output = {
            "DIMENSIONS": {
                "model": {dataset.attrs["source_id"]: {}},
                "region": {"global": {}},
                "metric": {"tas": {}},
                "json_structure": [
                    "model",
                    "region",
                    "metric",
                ],
            },
            "RESULTS": {
                dataset.attrs["source_id"]: {"global": {"tas": 0}},
            },
        }

        return cmec_output



from abc import abstractmethod
from collections.abc import Iterable
from pathlib import Path
from typing import ClassVar

import pandas
from ruamel.yaml import YAML

from cmip_ref_core.datasets import SourceDatasetType
from cmip_ref_core.metrics import CommandLineMetric, MetricExecutionDefinition, MetricExecutionResult
from cmip_ref_core.pycmec.metric import CMECMetric
from cmip_ref_core.pycmec.output import CMECOutput
from cmip_ref_metrics_esmvaltool.recipe import load_recipe, prepare_climate_data
from cmip_ref_metrics_esmvaltool.types import OutputBundle, Recipe

yaml = YAML()


class ESMValToolMetric(CommandLineMetric):
    """ESMValTool Metric base class."""

    base_recipe: ClassVar[str]

    @staticmethod
    @abstractmethod
    def update_recipe(recipe: Recipe, input_files: pandas.DataFrame) -> None:
        """
        Update the base recipe for the run.

        Parameters
        ----------
        recipe:
            The base recipe to update.
        input_files:
            The dataframe describing the input files.

        """

    @staticmethod
    @abstractmethod
    def format_result(result_dir: Path) -> OutputBundle:
        """
        Create a CMEC output bundle for the results.

        Parameters
        ----------
        result_dir
            Directory containing results from an ESMValTool run.

        Returns
        -------
            A CMEC output bundle.
        """

    def build_cmd(self, definition: MetricExecutionDefinition) -> Iterable[str]:
        """
        Build the command to run an ESMValTool recipe.

        Parameters
        ----------
        definition
            A description of the information needed for this execution of the metric

        Returns
        -------
        :
            The result of running the metric.
        """
        input_files = definition.metric_dataset[SourceDatasetType.CMIP6].datasets
        recipe = load_recipe(self.base_recipe)
        self.update_recipe(recipe, input_files)

        recipe_path = definition.to_output_path("recipe.yml")
        with recipe_path.open("w", encoding="utf-8") as file:
            yaml.dump(recipe, file)

        climate_data = definition.to_output_path("climate_data")

        prepare_climate_data(
            definition.metric_dataset[SourceDatasetType.CMIP6].datasets,
            climate_data_dir=climate_data,
        )

        config = {
            "drs": {
                "CMIP6": "ESGF",
            },
            "output_dir": str(definition.to_output_path("results")),
            "rootpath": {
                "default": str(climate_data),
            },
            "search_esgf": "never",
        }
        config_dir = definition.to_output_path("config")
        config_dir.mkdir()
        with (config_dir / "config.yml").open("w", encoding="utf-8") as file:
            yaml.dump(config, file)

        return [
            "esmvaltool",
            "run",
            f"--config-dir={config_dir}",
            f"{recipe_path}",
        ]

    def build_metric_result(
        self,
        definition: MetricExecutionDefinition,
    ) -> MetricExecutionResult:
        """
        Build the metric result after running an ESMValTool recipe.

        Parameters
        ----------
        definition
            A description of the information needed for this execution of the metric

        Returns
        -------
        :
            The resulting metric.
        """
        result_dir = next(definition.to_output_path("results").glob("*"))

        metric_bundle = self.format_result(result_dir)
        CMECMetric.model_validate(metric_bundle)

        output_bundle = CMECOutput.create_template()
        CMECOutput.model_validate(output_bundle)

        return MetricExecutionResult.build_from_output_bundle(
            definition,
            cmec_output_bundle=output_bundle,
            cmec_metric_bundle=metric_bundle,
        )



import importlib.metadata

from cmip_ref_metrics_esmvaltool import __version__, provider


def test_provider():
    assert provider.name == "ESMValTool"
    assert provider.slug == "esmvaltool"
    assert provider.version == __version__

    metric_modules = importlib.resources.files("cmip_ref_metrics_esmvaltool").glob("metrics/*.py")
    ignore = {
        "__init__.py",
        "base.py",
    }
    n_metric_modules = len([f for f in metric_modules if f.name not in ignore])
    assert len(provider) == n_metric_modules



from pathlib import Path

import numpy as np
import pandas
import xarray as xr
from cmip_ref_metrics_esmvaltool.metrics import TransientClimateResponse
from cmip_ref_metrics_esmvaltool.recipe import load_recipe


def test_update_recipe():
    input_files = pandas.read_json(Path(__file__).parent / "input_files_tcr.json")
    recipe = load_recipe("recipe_tcr.yml")
    TransientClimateResponse().update_recipe(recipe, input_files)
    assert len(recipe["datasets"]) == 2
    assert len(recipe["diagnostics"]) == 1
    assert set(recipe["diagnostics"]["cmip6"]["variables"]) == {"tas"}


def test_format_output(tmp_path):
    tcr = xr.Dataset(
        data_vars={
            "tcr": (["dim0"], np.array([1.0], dtype=np.float32)),
        },
        coords={
            "dataset": ("dim0", np.array([b"abc"])),
        },
    )
    result_dir = tmp_path
    subdir = result_dir / "work" / "cmip6" / "tcr"
    subdir.mkdir(parents=True)
    tcr.to_netcdf(subdir / "tcr.nc")

    output_bundle = TransientClimateResponse().format_result(result_dir)

    assert isinstance(output_bundle, dict)
    assert output_bundle["RESULTS"]["abc"]["global"]["tcr"] == 1.0



"""
Rapid evaluating CMIP data with ESMValTool.
"""

import cmip_ref_metrics_esmvaltool.metrics
from cmip_ref_core.providers import CondaMetricsProvider
from cmip_ref_metrics_esmvaltool._version import __version__

# Initialise the metrics manager and register the metrics
provider = CondaMetricsProvider("ESMValTool", __version__)
for _metric_cls_name in cmip_ref_metrics_esmvaltool.metrics.__all__:
    _metric_cls = getattr(cmip_ref_metrics_esmvaltool.metrics, _metric_cls_name)
    provider.register(_metric_cls())
